mod hiddenprotocol;
mod noroom;

use crate::{hiddenprotocol::*, noroom::*};
use cairo_sys::*;
use gdk_pixbuf_sys::*;
use gdk_sys::*;
use gio_sys::*;
use glib_sys::*;
use gobject_sys::*;
use gtk_sys::*;
use maraiah::{c_str, durandal::ffi};

unsafe extern "C" fn app_activate(app: *mut GtkApplication, dat: gpointer)
{
   let dat = dat as *mut Option<MapEditorState>;
   let dat = &mut *dat;
   let _ = dat;

   setup_css();

   let path = c_str!("/net/greyserv/maraiah/tycho/ui");
   let b    = gtk_builder_new_from_resource(path);

   setup_draw_area(b);
   setup_win_map_view(b);
   setup_win_map_tools(b);
   setup_win_map_prop(b);
   setup_about_dlg(b);
   setup_win_main(b, app);

   g_object_unref(b as _);
}

unsafe fn setup_draw_area(b: *mut GtkBuilder)
{
   struct RenderState
   {
      im_nomap: *mut GdkPixbuf,
      ax:       *mut GtkAdjustment,
      ay:       *mut GtkAdjustment,
   }

   unsafe extern "C" fn c_done(_: *mut GtkWidget, dat: gpointer)
   {
      let dat = Box::from_raw(dat as *mut RenderState);

      // unref everything
      g_object_unref(dat.im_nomap as _);
      g_object_unref(dat.ax as _);
      g_object_unref(dat.ay as _);

      // data is dropped and freed here
   }

   unsafe extern "C" fn c_draw(wid: *mut GtkWidget,
                               ctx: *mut cairo_t,
                               dat: gpointer)
      -> gboolean
   {
      let dat = dat as *mut RenderState;
      let dat = &mut *dat;

      let w = f64::from(gtk_widget_get_allocated_width(wid));
      let h = f64::from(gtk_widget_get_allocated_height(wid));

      gtk_adjustment_set_lower(dat.ax, 0.0);
      gtk_adjustment_set_upper(dat.ax, w);

      gtk_adjustment_set_lower(dat.ay, 0.0);
      gtk_adjustment_set_upper(dat.ay, h);

      let im = CrImage(dat.im_nomap);
      let dr = CrDrawArea::new(ctx, w, h);

      draw_map_none(&dr, &im);

      1
   }

   let wid: *mut GtkDrawingArea = get_obj(b, c_str!("draw-area"));

   let ax: *mut GtkAdjustment = get_obj(b, c_str!("adj-map-horz"));
   let ay: *mut GtkAdjustment = get_obj(b, c_str!("adj-map-vert"));

   g_object_ref(ax as _);
   g_object_ref(ay as _);

   let im_nomap = load_img(c_str!("/net/greyserv/maraiah/tycho/tycho1.png"));

   let dat = RenderState{im_nomap, ax, ay};
   let dat = Box::into_raw(Box::new(dat));

   connect(wid as _, c_str!("destroy"), c_done as _, dat as _);
   connect(wid as _, c_str!("draw"), c_draw as _, dat as _);
}

unsafe fn setup_win_map_view(b: *mut GtkBuilder)
{
   let win: *mut GtkWindow   = get_obj(b, c_str!("win-map-view"));
   let btn: *mut GtkMenuItem = get_obj(b, c_str!("btn-show-map-view"));

   connect_hide(win as _);
   connect_show(btn as _, win as _);
}

unsafe fn setup_win_map_tools(b: *mut GtkBuilder)
{
   let win: *mut GtkWindow   = get_obj(b, c_str!("win-map-tools"));
   let btn: *mut GtkMenuItem = get_obj(b, c_str!("btn-show-map-tools"));

   connect_hide(win as _);
   connect_show(btn as _, win as _);
}

unsafe fn setup_win_map_prop(b: *mut GtkBuilder)
{
   let win: *mut GtkWindow   = get_obj(b, c_str!("win-map-prop"));
   let btn: *mut GtkMenuItem = get_obj(b, c_str!("btn-show-map-prop"));

   connect_hide(win as _);
   connect_show(btn as _, win as _);
}

unsafe fn setup_about_dlg(b: *mut GtkBuilder)
{
   unsafe extern "C" fn c_show(_: *mut GtkWidget, dlg: gpointer)
   {
      gtk_dialog_run(dlg as _);
      gtk_widget_hide(dlg as _);
   }

   let dlg: *mut GtkAboutDialog = get_obj(b, c_str!("win-about"));
   let btn: *mut GtkMenuItem    = get_obj(b, c_str!("btn-about"));

   let it    = env!("CARGO_PKG_AUTHORS").split(';');
   let mut v = ffi::CStringVec::new_from_iter(it).unwrap();
   let img   = load_img(c_str!("/net/greyserv/maraiah/tycho/tycho2.png"));

   gtk_about_dialog_set_authors(dlg, v.as_mut_ptr());
   gtk_about_dialog_set_version(dlg, c_str!(env!("CARGO_PKG_VERSION")));
   gtk_about_dialog_set_website(dlg, c_str!(env!("CARGO_PKG_HOMEPAGE")));
   gtk_about_dialog_set_logo(dlg, img);

   connect_hide(dlg as _);
   connect(btn as _, c_str!("activate"), c_show as _, dlg as _);

   g_object_unref(img as _);
}

unsafe fn setup_win_main(b: *mut GtkBuilder, app: *mut GtkApplication)
{
   unsafe extern "C" fn c_quit(_: *mut GtkWidget, win: gpointer)
   {
      gtk_window_close(win as _);
   }

   let win: *mut GtkWindow   = get_obj(b, c_str!("win-main"));
   let btn: *mut GtkMenuItem = get_obj(b, c_str!("btn-quit"));

   gtk_window_set_application(win, app);
   gtk_widget_show_all(win as _);

   connect(btn as _, c_str!("activate"), c_quit as _, win as _);
}

unsafe fn setup_css()
{
   let path = c_str!("/net/greyserv/maraiah/tycho/css");
   let css  = gtk_css_provider_new();
   gtk_css_provider_load_from_resource(css, path);

   let scr = gdk_screen_get_default();
   let pri = GTK_STYLE_PROVIDER_PRIORITY_APPLICATION as u32;
   gtk_style_context_add_provider_for_screen(scr, css as _, pri);

   g_object_unref(css as _);
}

unsafe fn connect_hide(wid: *mut GtkWidget)
{
   unsafe extern "C" fn c_hide(wid: *mut GtkWidget,
                               _: *mut GdkEvent,
                               _: gpointer)
   {
      gtk_widget_hide(wid);
   }

   connect(wid as _, c_str!("delete-event"), c_hide as _, ffi::null_mut());
}

unsafe fn connect_show(btn: *mut GtkWidget, wid: *mut GtkWidget)
{
   unsafe extern "C" fn c_show(_: *mut GtkWidget, wid: gpointer)
   {
      gtk_widget_show_all(wid as _);
   }

   connect(btn as _, c_str!("activate"), c_show as _, wid as _);
}

/// Gets an object from a `GtkBuilder`.
unsafe fn get_obj<T>(b: *mut GtkBuilder, name: ffi::NT) -> *mut T
{
   let obj = gtk_builder_get_object(b, name);
   obj as _
}

/// Connects a signal handler.
unsafe fn connect(obj: *mut GObject, name: ffi::NT, cb: gpointer, d: gpointer)
{
   let cb = std::mem::transmute(cb);
   g_signal_connect_data(obj, name, cb, d, None, 0);
}

unsafe fn load_img(path: ffi::NT) -> *mut GdkPixbuf
{
   gdk_pixbuf_new_from_resource(path, ffi::null_mut())
}

/// Entry point.
fn main()
{
   unsafe {
      // get jacked, punk. opaque data structures are for nerds.
      const RESOURCE_DATA: &[u8] =
         include_bytes!(concat!(env!("OUT_DIR"), "/resources"));

      // first we create the static resource header, which is really simple
      let mut resource = GStaticResource{data:     RESOURCE_DATA.as_ptr(),
                                         data_len: RESOURCE_DATA.len(),
                                         resource: ffi::null_mut(),
                                         next:     ffi::null_mut(),
                                         padding:  ffi::null_mut()};

      // init it, now we can use it throughout the entire app without copying!
      g_static_resource_init(&mut resource);

      // create a container for the editor state
      let dat = Box::into_raw(Box::new(None::<MapEditorState>));

      // create and run the app
      let name = c_str!("net.greyserv.maraiah.tycho");
      let app = gtk_application_new(name, 0);

      connect(app as _, c_str!("activate"), app_activate as _, dat as _);

      g_application_run(app as _, 0, ffi::null_mut());

      // ok, clean up all this crap now
      g_object_unref(app as _);

      drop(Box::from_raw(dat));

      // deinit the "static" data, and everything will be done
      g_static_resource_fini(&mut resource);
   }
}

// EOF
